/* jack_diplomat - Joins jack daemons together
 * Copyright (C) 2004 Dave Cunningham <spark@xdev.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program (see COPYING); if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */



#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <float.h>
#include <limits.h>
#include <math.h>
#include <string.h>

#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <sys/mman.h>


#include <getopt.h>

#include <jack/jack.h>
#include <jack/ringbuffer.h>


#include "simple_src.h"
#include "rt_msg_buf.h"



#define VERSION "v0.70"



/* if this gets changed, also change the char[] in jack_diplomat_server_init */
#define MAX_CHANNELS 24
#define CALIBRATION_LOOPS 1000000








typedef struct jack_diplomat JackDiplomat;
int process (jack_nframes_t nframes, void *arg);
void jack_shutdown (void *arg);


typedef jack_default_audio_sample_t Smpl;


typedef struct jack_diplomat_server {

        jack_client_t *server;
        int server_opened;

        char *server_name;

        unsigned int in_ch, out_ch;

        jack_port_t *in[MAX_CHANNELS];

        jack_port_t *out[MAX_CHANNELS];

        jack_nframes_t  period;

        SimpleSRC *ssrc[MAX_CHANNELS];

        jack_ringbuffer_t *rbuf[MAX_CHANNELS];

        jack_nframes_t tol;
        jack_nframes_t buf_used;

        unsigned int no_report_xrun_periods;

        double period_avg;
        unsigned int period_avg_count;
        struct timespec ts_old;
        int running;

        jack_nframes_t log_length;

        double *min_log;
        int min_log_p;
        double min_log_run;

        double *max_log;
        int max_log_p;
        double max_log_run;

        double max_shunt;
        double ideal_shunt;
        double effective_shunt;
        double shunt_rate;

        JackDiplomat *root;

        jack_nframes_t log_counter;

        RTMsgBuf *buf;

} JackDiplomatServer;



int jack_diplomat_server_init(JackDiplomatServer *this,
                               char *client_name,
                               char *server_name,
                               unsigned int num_input,
                               unsigned int num_output,
                               double shunt_rate,
                               double max_shunt)
{

        jack_status_t status = 0;

        this->server = jack_client_open (client_name,
                                         JackServerName | JackNoStartServer |
                                                JackUseExactName,
                                         &status,
                                         server_name);

        if (!this->server) {
                fprintf(stderr,"Error: Failed to attach to jack server "
                               "\"%s\" as \"%s\"\n", server_name, client_name);

                if (status & JackFailure)
                        fprintf(stderr,"Overall operation failed.\n");

                if (status & JackInvalidOption)
                        fprintf(stderr, "The operation contained "
                                        " an invalid or unsupported option.\n");

                if (status & JackNameNotUnique)
                        fprintf(stderr,
                                "The desired client name was not unique.\n");

                if (status & JackServerFailed)
                        fprintf(stderr,
                                "Unable to connect to the JACK server.\n");

                if (status & JackServerError)
                        fprintf(stderr,
                                "Communication error with the JACK server.\n");

                if (status & JackNoSuchClient)
                        fprintf(stderr, "Requested client does not exist.\n");

                if (status & JackLoadFailure)
                        fprintf(stderr, "Unable to load internal client.\n");

                if (status & JackInitFailure)
                        fprintf(stderr, "Unable to initialize client.\n");

                if (status & JackShmFailure)
                        fprintf(stderr, "Unable to access shared memory.\n");

                if (status & JackVersionError)
                        fprintf(stderr,
                                "Client's protocol version does not match.\n");

                return -1;
        }

        this->server_opened = 1;

        this->server_name = server_name;

        jack_set_process_callback (this->server, process, this);
        jack_on_shutdown (this->server, jack_shutdown, this);


        assert(num_input<=MAX_CHANNELS);

        this->in_ch = num_input;
        this->out_ch= num_output;
        char port_name[20] = "";

        for (unsigned int i=0 ; i<num_input ; i++) {
                snprintf(port_name,20,"in_%d",i+1);
                this->in[i] = jack_port_register (this->server,
                                                  port_name,
                                                  JACK_DEFAULT_AUDIO_TYPE,
                                                  JackPortIsInput,
                                                  0);
                if (!this->in[i]) {
                        fprintf(stderr, "Error: Couldn't register port %s with "
                                        "jack daemon %s\n",
                                        port_name, server_name);
                        return -1;
                }
        }

        for (unsigned int i=0 ; i<num_output ; i++) {
                snprintf(port_name,20,"out_%d",i+1);
                this->out[i] = jack_port_register (this->server,
                                                   port_name,
                                                   JACK_DEFAULT_AUDIO_TYPE,
                                                   JackPortIsOutput,
                                                   0);
                if (!this->out[i]) {
                        fprintf(stderr, "Error: Couldn't register port %s with "
                                        "jack daemon %s\n",
                                        port_name, server_name);
                        return -1;
                }
        }

        this->period = jack_get_buffer_size(this->server);

        this->max_shunt = max_shunt;
        this->ideal_shunt = 0.0;
        this->effective_shunt = 0.0;
        this->shunt_rate = shunt_rate;

        this->log_counter = 0;

        return 0;
}

int jack_diplomat_server_size(JackDiplomatServer *this,
                               JackDiplomatServer *other,
                               jack_nframes_t tol)
{

        jack_nframes_t max_period = this->period > other->period ?
                                   this->period : other->period;
        jack_nframes_t min_period = this->period > other->period ?
                                   other->period : this->period;

        this->tol = tol;
        jack_nframes_t cap = min_period + max_period + min_period;
        for (unsigned int i=0 ; i<this->in_ch ; i++) {
                this->rbuf[i] = jack_ringbuffer_create(
                        sizeof(Smpl) * (tol + cap + tol));
                this->ssrc[i] = simple_src_new();
        }
        this->buf_used = 0;

        this->no_report_xrun_periods = 20; //1+other->period / this->period;

        this->period_avg = 0.0;
        this->period_avg_count = 0.0;
        this->running = 0;

        this->log_length = (max_period+min_period-1)/min_period+1;

        this->min_log = malloc(sizeof(*this->min_log) * this->log_length);

        this->max_log = malloc(sizeof(*this->max_log) * this->log_length);

        this->buf = rt_msg_buf_new(1000000);

        if (!this->buf) {
                fprintf(stderr,"Could not allocated an RTMsgBuf instance\n");
                return -1;
        }

        if (!this->min_log || !this->max_log) {
                fprintf(stderr,"Could not allocate memory for the tiny buffers "
                               "used to calculate statistics for the dynamic "
                               "frequency correction.\n");
                return -1;
        }

        for (unsigned int i=0 ; i<this->in_ch ; i++) {
                if (!this->rbuf[i]) {
                        fprintf(stderr,"Error: Could not allocate buffer for "
                                       "out%d coming from jack server %s.\n",
                                       i, this->server_name);
                        return -1;
                }
                if (!this->ssrc[i]) {
                        fprintf(stderr,"Error: Could not allocate a new sample "
                                       "rate convertor for out%d coming from "
                                       "jack server %s.\n",
                                       i, this->server_name);
                        return -1;
                }
        }

        for (jack_nframes_t i=0 ; i<this->log_length ; i++)
                this->min_log[i] = FLT_MAX;
        this->min_log_p = 0;
        this->min_log_run = FLT_MAX;

        for (jack_nframes_t i=0 ; i<this->log_length ; i++)
                this->max_log[i] = FLT_MIN;
        this->max_log_p = 0;
        this->max_log_run = FLT_MIN;

        return 0;
}



void jack_diplomat_server_del (JackDiplomatServer *this)
{
        if (this->server_opened) jack_client_close (this->server);

        rt_msg_buf_del(this->buf);

}






struct jack_diplomat {

        JackDiplomatServer s[2];

        int quit;

        int begin;

        int verbose;

};



int jack_diplomat_init(JackDiplomat *this,
                        char *client_name1,
                        char *client_name2,
                        char *server_name1,
                        char *server_name2,
                        unsigned int forwards,
                        unsigned int backwards,
                        jack_nframes_t tol,
                        double shunt_rate,
                        double max_shunt,
                        int verbose
                        )
{
        this->quit = 0;

        this->s[0].root = this;
        this->s[1].root = this;

        this->verbose = verbose;

        int error;

        error = jack_diplomat_server_init(&this->s[0],
                                          client_name1,
                                          server_name1,
                                          forwards,
                                          backwards,
                                          shunt_rate,max_shunt);
        if (error) return error;

        error = jack_diplomat_server_init(&this->s[1],
                                          client_name2,
                                          server_name2,
                                          backwards,
                                          forwards,
                                          shunt_rate,max_shunt);
        if (error) return error;

        error = jack_diplomat_server_size(&this->s[0], &this->s[1], tol);
        if (error) return error;

        error = jack_diplomat_server_size(&this->s[1], &this->s[0], tol);
        if (error) return error;

        return 0;
}



void jack_diplomat_del (JackDiplomat *this)
{
        jack_diplomat_server_del (&this->s[0]);
        jack_diplomat_server_del (&this->s[1]);
}






void update_stats(JackDiplomatServer *this, JackDiplomatServer *other)
{
        double victim;

        /* if the new value is bigger than the logged values, set max to it */
        if (this->buf_used > this->max_log_run)
                this->max_log_run = this->buf_used;

        victim = this->max_log[this->max_log_p];
        this->max_log[this->max_log_p] = this->buf_used;
        this->max_log_p = (this->max_log_p + 1) % this->log_length;

        /* alternatively if the removed value was a maximum, recalculate max */
        if (victim == this->max_log_run) {
                this->max_log_run = INT_MIN;
                for (unsigned int i=0 ; i<this->log_length ; i++) {
                        if (this->max_log[i] > this->max_log_run)
                                this->max_log_run = this->max_log[i];
                }
        }


        if (this->buf_used < this->min_log_run)
                this->min_log_run = this->buf_used;

        victim = this->min_log[this->min_log_p];
        this->min_log[this->min_log_p] = this->buf_used;
        this->min_log_p = (this->min_log_p + 1) % this->log_length;

        if (victim == this->min_log_run) {
                this->min_log_run = INT_MAX;
                for (unsigned int i=0 ; i<this->log_length ; i++) {
                        if (this->min_log[i] < this->min_log_run)
                                this->min_log_run = this->min_log[i];
                }
        }


        double middle = (this->min_log_run + this->max_log_run)/2;

        jack_nframes_t max_frame = this->period > other->period ?
                                   this->period : other->period;
        jack_nframes_t min_frame = this->period > other->period ?
                                   other->period : this->period;

/* not sure about this anymore
        double equator = this->tol +
                         (this->period_avg>other->period_avg?min_frame:0) +
                         max_frame/2;
*/

        double equator = this->tol + min_frame + max_frame/2;

        //this->ideal_shunt = (middle>equator?-1.0:1.0) * this->max_shunt;

        this->ideal_shunt = - this->max_shunt * (middle-equator) / middle;

        if (this->root->verbose>=3)
                rt_msg_buf_printf(this->buf,"%s Max: %f\n",this->server_name,
                                                           this->max_log_run);

        if (this->root->verbose>=3)
                rt_msg_buf_printf(this->buf,"%s Min: %f\n",this->server_name,
                                                           this->min_log_run);

        if (this->root->verbose>=2)
                rt_msg_buf_printf(this->buf,"%s Middle: %f\n",this->server_name,
                                                              middle);

        if (this->root->verbose>=1)
                rt_msg_buf_printf(this->buf,"%s BufferFill: %d\n",
                                            this->server_name, this->buf_used);
}



void update_shunt(JackDiplomatServer *this)
{
        if (this->effective_shunt > this->ideal_shunt) {

                this->effective_shunt -= this->shunt_rate;

                if (this->effective_shunt < this->ideal_shunt)
                        this->effective_shunt = this->ideal_shunt;

        } else if (this->effective_shunt < this->ideal_shunt) {

                this->effective_shunt += this->shunt_rate;

                if (this->effective_shunt > this->ideal_shunt)
                        this->effective_shunt = this->ideal_shunt;

        }
}






int process (jack_nframes_t nframes, void *arg)
{
        JackDiplomatServer *s = (JackDiplomatServer*) arg;
        JackDiplomat *this = s->root;
        JackDiplomatServer *s2 = &this->s[0] == s ? &this->s[1] : &this->s[0];

        if (s->period_avg_count*nframes<CALIBRATION_LOOPS) {
                /* spend the first s->period_avg_count cycles calculating 
                 * the average period length in clock cycles */
                static struct timespec ts_new;


                clock_gettime(CLOCK_REALTIME, &ts_new);
                if (s->running) {
                        long int diff = ts_new.tv_nsec - s->ts_old.tv_nsec;
                        diff = diff<0 ? diff+1E9 : diff;
                        s->period_avg*=s->period_avg_count++;
                        s->period_avg+=(diff)/(double)nframes;
                        s->period_avg/=s->period_avg_count;
                }
                s->ts_old = ts_new;
                s->running = 1;

        } else if (s->period_avg_count*nframes>=CALIBRATION_LOOPS &&
                   s2->period_avg_count*s2->period>=CALIBRATION_LOOPS &&
                   !s->root->quit) {

                double calibrated_ratio = s2->period_avg/s->period_avg;

                if (calibrated_ratio<0.99 || calibrated_ratio > 1.01) {
                        rt_msg_buf_printf(s->buf,
                                "Cannot correct frequency by this much %f\n",
                                          calibrated_ratio);
                        s->root->quit = 1;
                        return 0;
                }

                jack_nframes_t pushed_data = 0, pulled_data = 0;

                jack_nframes_t overrun;

                for (unsigned int i=0 ; i<s->in_ch; i++) {

                        Smpl *from_jack= jack_port_get_buffer(s->in[i],nframes);

                        jack_ringbuffer_data_t vec[2];
                        jack_ringbuffer_get_write_vector (s->rbuf[i], vec);

                        if (this->verbose >= 5)
                                rt_msg_buf_printf(s->buf,"%s: ringbuf %d %d\n",
                                          s->server_name,
                                          vec[0].len / sizeof(Smpl),
                                          vec[1].len / sizeof(Smpl));


                        pushed_data = simple_src_process(
                                        s->ssrc[i],
                                        from_jack,
                                        nframes,
                                        (Smpl*)vec[0].buf,
                                        vec[0].len / sizeof(Smpl),
                                        calibrated_ratio-s->effective_shunt,
                                        &pulled_data,
                                        NULL);

                        assert(pulled_data <= nframes);

                        if (this->verbose >= 4)
                                rt_msg_buf_printf(s->buf,"%s: pulled1: %d\n",
                                                 s->server_name, pulled_data);

                        // if pulled_data tells us everything has been taken
                        // from this jack frame then we do not need to send the
                        // rest to the next vector

                        overrun = 0; // reset from previous channel


                        if (pulled_data < nframes) {
                                pushed_data += simple_src_process(
                                        s->ssrc[i],
                                        from_jack + pulled_data,
                                        nframes - pulled_data,
                                        (Smpl*)vec[1].buf,
                                        (vec[1].len/sizeof(Smpl)),
                                        calibrated_ratio-s->effective_shunt,
                                        &pulled_data,
                                        &overrun);

                                if (this->verbose >= 4)
                                rt_msg_buf_printf(s->buf,"%s: pulled2: %d\n",
                                                 s->server_name, pulled_data);

                        }

                        // if overrun is 1 now, we overran the second buffer so
                        // an actual overrun has occured.

                        // pulled_data will also not be enough but since we did
                        // not record the amount for the first buffer, we cannot
                        // add it on to get the total proportion of nframes
                        // consumed.  we lose data and reset the stream anyway
                        // so this does not matter

                        if (overrun) {
                                simple_src_reset(s->ssrc[i]);
                        }

                        jack_ringbuffer_write_advance (
                                        s->rbuf[i],
                                        pushed_data * sizeof(Smpl)) ;

                }


                if (s->in_ch > 0 &&
                    overrun &&
                    s->no_report_xrun_periods==0 &&
                    this->verbose>=0) {

                        rt_msg_buf_printf(s->buf,"%s OVERrun of: %d\n",
                                          s->server_name, overrun);
                }



                // only update the stats if we did anything
                if (s->in_ch > 0) {
                        // should be the same for all, so we use the last one
                        s->buf_used += pushed_data;
                        update_stats(s,s2);
                        update_shunt(s);
                }



                jack_nframes_t clipped_pop = 0;
                jack_nframes_t underrun = 0;

                for (unsigned int i=0 ; i<s->out_ch; i++) {

                        Smpl *to_jack =
                                jack_port_get_buffer (s->out[i], nframes);

                        if (s2->buf_used>=nframes) {
                                clipped_pop = nframes;
                        } else {
                                underrun = nframes - s2->buf_used;
                                clipped_pop = s2->buf_used > s2->tol ?
                                              s2->buf_used - s2->tol :
                                              s2->buf_used;
                        }

                        jack_ringbuffer_read (s2->rbuf[i],
                                              (char*)to_jack,
                                              sizeof(Smpl)*clipped_pop);

                }

                if (underrun && s->no_report_xrun_periods==0 &&this->verbose>=0)
                        rt_msg_buf_printf(s->buf,"%s UNDERrun of %d\n",
                                          s->server_name, underrun);

                s2->buf_used -= clipped_pop;

                if (underrun || clipped_pop) {
                        update_stats(s2,s);
                }

                if (s->no_report_xrun_periods>0)
                        s->no_report_xrun_periods--;

        }

        return 0;
}






JackDiplomat *state_ptr = NULL;






void signal_handler (int ignore_this_arg)
{
        (void) ignore_this_arg;
        state_ptr->quit = 1;
}






void error (const char *desc)
{
        fprintf (stderr, "JACK error: %s\n", desc);
}






void jack_shutdown (void *arg)
{
        JackDiplomatServer *this = (JackDiplomatServer*) arg; 
        fprintf(stderr,"Jack shut us down!\n");
        this->root->quit = 1;
}






void print_usage(char *name, FILE *to)
{
        /* I lost 20% in a university unix assignment once because I spelt it
         * "usage" instead of "Usage".  NEVER AGAIN!
         */
        fprintf(to,
                "%s "VERSION": Connects two jack servers together.\n",name);
        fprintf(to,
                "    (Author: Dave Cunningham <spark@xdev.net>  License:GPL)\n"
                "\n"
                "Usage: %s [OPTION]... -1 <server_name> -2 <server_name>\n"
                "\n"
                "        -h --help\tDisplay this usage info\n"
                "        -v --verbose\tMore status messages\n"
                "        -q --quiet\tFewer messages (except fatal errors)\n"
                "        -c --calquiet\tNo calibration progress info\n"
                "\n"
                "           --name1\tClient name for jack (first server)\n"
                "           --name2\tClient name for jack (second server)\n"
                "        -N --name\tClient name for jack (both servers)\n"
                "        -1 --sname1\tName of the first jack server\n"
                "        -2 --sname2\tName of the second jack server\n"
                "\n"
                "        -f --forwards\tNumber of forwards channels\n"
                "        -b --backwards\tNumber of backwards channels\n"
                "\n"
                "             (the following options tweak the buffer)\n"
                "\n"
                "        -t --tolerance\tAdditional buffer at top and bottom\n"
                "        -s --shunt\tMax dynamic freq correction\n"
                "        -S --shuntrate\tRate of change of freq correction\n"
                "        -r --ratio\tSpecify the samplerate ratio s1/s2\n"
                "\n"
                ,name);
        fprintf(to,
                "        Where command line options conflict with each other,\n"
                "        the earlier options are overridden.\n");
}






int main (int argc, char *argv[])
{

        char *app_name = strrchr(argv[0], '/');
        app_name = app_name==NULL ? argv[0] : app_name + 1;

        /* test that NULL is infact 0 */
        void *ptr;
        memset(&ptr,0,sizeof(ptr));
        if (ptr!=NULL) {
                fprintf(stderr,"This program assumes the NULL pointer is 0\n");
                exit(EXIT_FAILURE);
        }

        JackDiplomat state;
        memset(&state,0,sizeof(state)); /* set all initial pointers to NULL */
        state_ptr = &state; /* global variable for signal handler */

        if(signal(SIGINT,signal_handler)==SIG_ERR) {
                perror("Registering SIGINT handler.");
                exit(EXIT_FAILURE);
        }

        const char *options = "hvqc" "N:1:2:" "f:b:" "t:s:S:r:";

        struct option long_options[] = {
                {"help", 0, 0, 'h'},
                {"verbose", 0, 0, 'v'},
                {"quiet", 0, 0, 'q'},
                {"calquiet", 0, 0, 'c'},

                {"sname1", 1, 0, '1'},
                {"sname2", 1, 0, '2'},
                {"name", 1, 0, 'N'},
                {"name1", 1, 0, '3'},
                {"name2", 1, 0, '4'},

                {"forwards", 1, 0, 'f'},
                {"backwards", 1, 0, 'b'},

                {"tolerance", 1, 0, 't'},
                {"shunt", 1, 0, 's'},
                {"shuntrate", 1, 0, 'S'},
                {"ratio", 1, 0, 'r'},
                {0, 0, 0, 0}
        };

        int opt = getopt_long(argc, argv, options, long_options, NULL);

        int verbose = 0, cal_quiet = 0;

        char *name1 = app_name, *name2 = app_name;
        char *sname1 = NULL, *sname2 = NULL;

        unsigned long forwards = 2, backwards = 2;

        jack_nframes_t tol = 10;

        double shunt = 0.0005, shuntrate = 0.000005, ratio=0.0;

        while (opt!=EOF) {
                switch (opt) {
                case 'h':
                        print_usage(app_name,stdout);
                        return EXIT_SUCCESS;
                case 'v':
                        verbose++;
                        break;
                case 'q':
                        verbose--;
                        break;
                case 'c':
                        cal_quiet = 1;
                        break;
                case '3':
                        name1 = optarg;
                        break;
                case '4':
                        name2 = optarg;
                        break;
                case 'N':
                        name1 = optarg;
                        name2 = optarg;
                        break;
                case '1':
                        if (!sname1) sname1 = optarg;
                        else if (!sname2) sname2 = optarg;
                        break;
                case '2':
                        if (!sname1) sname1 = optarg;
                        else if (!sname2) sname2 = optarg;
                        break;
                case 'f':
                        forwards = strtoul(optarg,NULL,10);
                        if (forwards > MAX_CHANNELS) {
                                fprintf(stderr,"No more than %d channels "
                                               "allowed.\n",MAX_CHANNELS);
                                forwards = MAX_CHANNELS;
                        }
                        break;
                case 'b':
                        backwards = strtoul(optarg,NULL,10);
                        if (backwards > MAX_CHANNELS) {
                                fprintf(stderr,"No more than %d channels "
                                               "allowed.\n",MAX_CHANNELS);
                                backwards = MAX_CHANNELS;
                        }
                        break;
                case 't':
                        tol = strtod(optarg,NULL);
                        break;
                case 's':
                        shunt = strtod(optarg,NULL);
                        break;
                case 'S':
                        shuntrate = strtod(optarg,NULL);
                        break;
                case 'r':
                        ratio = strtod(optarg,NULL);
                        break;
                default:
                        fprintf(stderr, "Unknown option: %c\n", optopt);
                        print_usage(app_name,stderr);
                        return EXIT_FAILURE;
                }
                opt = getopt_long(argc, argv, options, long_options, NULL);
        }

        if (!sname1 || !sname2) {
                fprintf(stderr, "Error: No server name specified.\n");
                fprintf(stderr, "Please use cmdline options -1 and -2.\n");
                fprintf(stderr, "See \"%s --help\" for more info.\n",app_name);
                //print_usage(app_name,stderr);
                exit(EXIT_FAILURE);
        }

        int error = mlockall(MCL_CURRENT || MCL_FUTURE);
        if (error) {
                jack_diplomat_del(&state);
                fprintf(stderr,"Failed to mlock memory %s.\n",app_name);
                exit(EXIT_FAILURE);
        }

        error = jack_diplomat_init(&state,
                                   name1,name2,sname1,sname2,
                                   forwards,backwards,
                                   tol,shuntrate,shunt,verbose);
        if (error) {
                jack_diplomat_del(&state);
                fprintf(stderr,"Failed to initialise %s.\n",app_name);
                exit(EXIT_FAILURE);
        }

        int completed_calibration = 0;

        if (ratio!=0.0) {

                state.s[0].period_avg = 1.0;
                state.s[0].period_avg_count = CALIBRATION_LOOPS;

                state.s[1].period_avg = ratio;
                state.s[1].period_avg_count = CALIBRATION_LOOPS;

                completed_calibration = 1;
        }


        jack_activate(state.s[0].server);
        jack_activate(state.s[1].server);


        struct timespec ts = {0, 5E7};

        while (!state.quit) {
                nanosleep (&ts,0);
                unsigned int a,b;
                a = state.s[0].period_avg_count*state.s[0].period;
                b = state.s[1].period_avg_count*state.s[1].period;
                unsigned int min = a<b ? a : b;
                rt_msg_buf_poll(state.s[0].buf,stdout);
                rt_msg_buf_poll(state.s[1].buf,stdout);
                if (!completed_calibration && !cal_quiet) {
                        printf("\rCalibration %d%% complete",
                                (100*min)/CALIBRATION_LOOPS);
                        if (min>=CALIBRATION_LOOPS) {
                                completed_calibration=1;
                                printf("\rCalibration finished        \n");
                                printf("The ratio is: %.15f\n",
                                       state.s[1].period_avg /
                                       state.s[0].period_avg);
                        }
                }

                fflush(stdout);
                fflush(stderr);
        }

        rt_msg_buf_poll(state.s[0].buf,stdout);
        rt_msg_buf_poll(state.s[1].buf,stdout);

        jack_diplomat_del (&state);

        return EXIT_SUCCESS;
}
