/* jack_timing - Gathers timing info from a pair of jack servers
 * Copyright (C) 2004 Dave Cunningham <spark@xdev.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program (see COPYING); if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */



#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <float.h>
#include <limits.h>
#include <math.h>
#include <string.h>

#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <sys/mman.h>
#include <pthread.h>


#include <getopt.h>

#include <jack/jack.h>
#include <jack/ringbuffer.h>


#include "rt_msg_buf.h"



#define JACK_TIMING_VERSION "v0.50"











typedef struct jack_timing JackTiming;
int process (jack_nframes_t nframes, void *arg);
void jack_shutdown (void *arg);


typedef jack_default_audio_sample_t Smpl;


typedef struct jack_timing_server {

        jack_client_t *server;
        int server_opened;

        char *server_name;

        jack_nframes_t period;

        struct timespec ts_old;
        int running;

        JackTiming *root;

        RTMsgBuf *buf;

        char score;

        pthread_mutex_t mutex;

} JackTimingServer;



int jack_timing_server_init(JackTimingServer *this,
                               char *client_name,
                               char *server_name)
{

        this->server_opened = 0;

        jack_status_t status = 0;

        this->server = jack_client_open (client_name,
                                         JackServerName | JackNoStartServer |
                                                JackUseExactName,
                                         &status,
                                         server_name);

        if (!this->server) {
                fprintf(stderr,"Error: Failed to attach to jack server "
                               "\"%s\" as \"%s\"\n", server_name, client_name);

                if (status & JackFailure)
                        fprintf(stderr,"Overall operation failed.\n");

                if (status & JackInvalidOption)
                        fprintf(stderr, "The operation contained "
                                        " an invalid or unsupported option.\n");

                if (status & JackNameNotUnique)
                        fprintf(stderr,
                                "The desired client name was not unique.\n");

                if (status & JackServerFailed)
                        fprintf(stderr,
                                "Unable to connect to the JACK server.\n");

                if (status & JackServerError)
                        fprintf(stderr,
                                "Communication error with the JACK server.\n");

                if (status & JackNoSuchClient)
                        fprintf(stderr, "Requested client does not exist.\n");

                if (status & JackLoadFailure)
                        fprintf(stderr, "Unable to load internal client.\n");

                if (status & JackInitFailure)
                        fprintf(stderr, "Unable to initialize client.\n");

                if (status & JackShmFailure)
                        fprintf(stderr, "Unable to access shared memory.\n");

                if (status & JackVersionError)
                        fprintf(stderr,
                                "Client's protocol version does not match.\n");

                return -1;
        }

        this->server_opened = 1;

        this->server_name = server_name;

        jack_set_process_callback (this->server, process, this);
        jack_on_shutdown (this->server, jack_shutdown, this);


        this->period = jack_get_buffer_size(this->server);

        this->score = 0;

        this->running = 0;

        this->buf = rt_msg_buf_new(1000000);

        if (!this->buf) {
                fprintf(stderr,"Could not allocated an RTMsgBuf instance\n");
                return -1;
        }

        int error = pthread_mutex_init(&this->mutex, NULL);
        if (error) {
                fprintf(stderr,"Failed to initialise mutex for server %s.\n",
                        this->server_name);
                exit(EXIT_FAILURE);
        }


        return 0;
}



void jack_timing_server_del (JackTimingServer *this)
{
        if (this->server_opened) jack_client_close (this->server);

        rt_msg_buf_del(this->buf);

}






struct jack_timing {

        JackTimingServer s[2];

        int quit;

        int begin;

        int verbose;

};



int jack_timing_init(JackTiming *this,
                        char *client_name1,
                        char *client_name2,
                        char *server_name1,
                        char *server_name2,
                        int verbose
                        )
{
        this->quit = 0;

        this->s[0].root = this;
        this->s[1].root = this;

        this->verbose = verbose;

        int error;

        error = jack_timing_server_init(&this->s[0],
                                          client_name1,
                                          server_name1);
        if (error) return error;

        error = jack_timing_server_init(&this->s[1],
                                          client_name2,
                                          server_name2);
        if (error) return error;

        return 0;
}



void jack_timing_del (JackTiming *this)
{
        jack_timing_server_del (&this->s[0]);
        jack_timing_server_del (&this->s[1]);
}






int process (jack_nframes_t nframes, void *arg)
{
        (void) nframes;
        JackTimingServer *s = (JackTimingServer*) arg;
        JackTiming *this = s->root;
        JackTimingServer *s2 = &this->s[0] == s ? &this->s[1] : &this->s[0];
        pthread_mutex_lock(&s->mutex);

        if (!this->quit) {

                // measuring period nanoseconds

                static struct timespec ts_new;


                clock_gettime(CLOCK_REALTIME, &ts_new);
                if (s->running) {
                        long int diff = ts_new.tv_nsec - s->ts_old.tv_nsec;
                        diff = diff<0 ? diff+1E9 : diff;
                        
                        //rt_msg_buf_printf(s->buf,"%s: time: %d %d\n",
                        //                  s->server_name,
                        //                  ts_new.tv_sec,
                        //                  ts_new.tv_nsec);
                        
                        if (diff>1E9*1.2*32/48000)
                                rt_msg_buf_printf(s->buf,
                                                  "%*lu.%09lu %s long! %f\n",
                                                  ts_new.tv_sec,
                                                  ts_new.tv_nsec,
                                                  s->server_name,
                                                  diff/32.0*48000*1E-9);
                        //rt_msg_buf_printf(s->buf,"%s: period_ns: %d\n",
                        //                  s->server_name, diff);
                }
                s->ts_old = ts_new;
                s->running = 1;




                s->score++;

                s2->score = 0;

                if (s->score>1)
                        rt_msg_buf_printf(s->buf,
                                          "%lu.%09lu %s cheated %d times\n",
                                          ts_new.tv_sec,
                                          ts_new.tv_nsec,
                                          s->server_name,
                                          s->score);
        }


        pthread_mutex_unlock(&s->mutex);
        return 0;
}






JackTiming *state_ptr = NULL;






void signal_handler (int ignore_this_arg)
{
        (void) ignore_this_arg;
        state_ptr->quit = 1;
}






void error (const char *desc)
{
        fprintf (stderr, "JACK error: %s\n", desc);
}






void jack_shutdown (void *arg)
{
        JackTimingServer *this = (JackTimingServer*) arg; 
        fprintf(stderr,"Jack shut us down!\n");
        this->root->quit = 1;
}






void print_usage(char *name, FILE *to)
{
        /* I lost 20% in a university unix assignment once because I spelt it
         * "usage" instead of "Usage".  NEVER AGAIN!
         */
        fprintf(to,
                "%s "JACK_TIMING_VERSION": Gathers timing info from a pair "
                    "of jack servers.\n",name);
        fprintf(to,
                "    (Author: Dave Cunningham <spark@xdev.net>  License:GPL)\n"
                "\n"
                "Usage: %s [OPTION]... -1 <server_name> -2 <server_name>\n"
                "\n"
                "        -h --help\tDisplay this usage info\n"
                "        -v --verbose\tMore status messages\n"
                "        -q --quiet\tFewer messages (except fatal errors)\n"
                "\n"
                "           --name1\tClient name for jack (first server)\n"
                "           --name2\tClient name for jack (second server)\n"
                "        -N --name\tClient name for jack (both servers)\n"
                "        -1 --sname1\tName of the first jack server\n"
                "        -2 --sname2\tName of the second jack server\n"
                "                   \t(can set both to the same server)\n"
                "\n"
                ,name);
        fprintf(to,
                "        Where command line options conflict with each other,\n"
                "        the earlier options are overridden.\n");
}






int main (int argc, char *argv[])
{

        char *app_name = strrchr(argv[0], '/');
        app_name = app_name==NULL ? argv[0] : app_name + 1;

        /* test that NULL is infact 0 */
        void *ptr;
        memset(&ptr,0,sizeof(ptr));
        if (ptr!=NULL) {
                fprintf(stderr,"This program assumes the NULL pointer is 0\n");
                exit(EXIT_FAILURE);
        }

        JackTiming state;
        memset(&state,0,sizeof(state)); /* set all initial pointers to NULL */
        state_ptr = &state; /* global variable for signal handler */

        if(signal(SIGINT,signal_handler)==SIG_ERR) {
                perror("Registering SIGINT handler.");
                exit(EXIT_FAILURE);
        }

        const char *options = "hvq" "N:1:2:";

        struct option long_options[] = {
                {"help", 0, 0, 'h'},
                {"verbose", 0, 0, 'v'},
                {"quiet", 0, 0, 'q'},

                {"sname1", 1, 0, '1'},
                {"sname2", 1, 0, '2'},
                {"name", 1, 0, 'N'},
                {"name1", 1, 0, '3'},
                {"name2", 1, 0, '4'},

                {0, 0, 0, 0}
        };

        int opt = getopt_long(argc, argv, options, long_options, NULL);

        int verbose = 0;

        char *name1 = app_name, *name2 = app_name;
        char *sname1 = NULL, *sname2 = NULL;

        while (opt!=EOF) {
                switch (opt) {
                case 'h':
                        print_usage(app_name,stdout);
                        return EXIT_SUCCESS;
                case 'v':
                        verbose++;
                        break;
                case 'q':
                        verbose--;
                        break;
                case '3':
                        name1 = optarg;
                        break;
                case '4':
                        name2 = optarg;
                        break;
                case 'N':
                        name1 = optarg;
                        name2 = optarg;
                        break;
                case '1':
                        if (!sname1) sname1 = optarg;
                        else if (!sname2) sname2 = optarg;
                        break;
                case '2':
                        if (!sname1) sname1 = optarg;
                        else if (!sname2) sname2 = optarg;
                        break;
                default:
                        fprintf(stderr, "Unknown option: %c\n", optopt);
                        print_usage(app_name,stderr);
                        return EXIT_FAILURE;
                }
                opt = getopt_long(argc, argv, options, long_options, NULL);
        }

        if (!sname1 || !sname2) {
                fprintf(stderr, "Error: No server name specified.\n");
                fprintf(stderr, "Please use cmdline options -1 and -2.\n");
                fprintf(stderr, "See \"%s --help\" for more info.\n",app_name);
                exit(EXIT_FAILURE);
        }

        int error = mlockall(MCL_CURRENT || MCL_FUTURE);
        if (error) {
                jack_timing_del(&state);
                fprintf(stderr,"Failed to mlock memory %s.\n",app_name);
                exit(EXIT_FAILURE);
        }

        error = jack_timing_init(&state, name1,name2,sname1,sname2, verbose);
        if (error) {
                jack_timing_del(&state);
                fprintf(stderr,"Failed to initialise %s.\n",app_name);
                exit(EXIT_FAILURE);
        }

        jack_activate(state.s[0].server);
        jack_activate(state.s[1].server);


        struct timespec ts = {0, 5E7};

        while (!state.quit) {
                nanosleep (&ts,0);
                rt_msg_buf_poll(state.s[0].buf,stdout);
                rt_msg_buf_poll(state.s[1].buf,stdout);
                fflush(stdout);
                fflush(stderr);
        }

        rt_msg_buf_poll(state.s[0].buf,stdout);
        rt_msg_buf_poll(state.s[1].buf,stdout);

        jack_timing_del (&state);

        return EXIT_SUCCESS;
}
