/* jack_timing - Gathers timing info from a pair of jack servers
 * Copyright (C) 2004 Dave Cunningham <spark@xdev.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program (see COPYING); if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */



#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <float.h>
#include <limits.h>
#include <math.h>
#include <string.h>

#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <sys/mman.h>
#include <pthread.h>


#include <getopt.h>

#include <jack/jack.h>
#include <jack/ringbuffer.h>


#include "rt_msg_buf.h"



#define JACK_TIMING_VERSION "v0.50"











int process (jack_nframes_t nframes, void *arg);
void jack_shutdown (void *arg);



typedef struct jack_timing {

        int quit;

        int begin;

        int verbose;

        jack_client_t *server;
        int server_opened;

        jack_nframes_t period;

        struct timespec ts_old;
        int running;

        RTMsgBuf *buf;

        char score;

} JackTiming;



int jack_timing_init(JackTiming  *this, char *client_name, int verbose)
{
        this->verbose = verbose;

        jack_status_t status = 0;

        this->server = jack_client_open (client_name,
                                         JackNoStartServer,
                                         &status);

        if (!this->server) {
                fprintf(stderr,"Error: Failed to attach to jack server "
                               "as \"%s\"\n", client_name);

                if (status & JackFailure)
                        fprintf(stderr,"Overall operation failed.\n");

                if (status & JackInvalidOption)
                        fprintf(stderr, "The operation contained "
                                        " an invalid or unsupported option.\n");

                if (status & JackNameNotUnique)
                        fprintf(stderr,
                                "The desired client name was not unique.\n");

                if (status & JackServerFailed)
                        fprintf(stderr,
                                "Unable to connect to the JACK server.\n");

                if (status & JackServerError)
                        fprintf(stderr,
                                "Communication error with the JACK server.\n");

                if (status & JackNoSuchClient)
                        fprintf(stderr, "Requested client does not exist.\n");

                if (status & JackLoadFailure)
                        fprintf(stderr, "Unable to load internal client.\n");

                if (status & JackInitFailure)
                        fprintf(stderr, "Unable to initialize client.\n");

                if (status & JackShmFailure)
                        fprintf(stderr, "Unable to access shared memory.\n");

                if (status & JackVersionError)
                        fprintf(stderr,
                                "Client's protocol version does not match.\n");

                return -1;
        }

        this->server_opened = 1;

        jack_set_process_callback (this->server, process, this);
        jack_on_shutdown (this->server, jack_shutdown, this);


        this->period = jack_get_buffer_size(this->server);

        this->score = 0;

        this->running = 0;

        this->buf = rt_msg_buf_new(1000000);

        if (!this->buf) {
                fprintf(stderr,"Could not allocated an RTMsgBuf instance\n");
                return -1;
        }


        return 0;
}








void jack_timing_del (JackTiming *this)
{
        if (this->server_opened) jack_client_close (this->server);

        rt_msg_buf_del(this->buf);
}






int process (jack_nframes_t nframes, void *arg)
{
        (void) nframes;
        JackTiming *this = (JackTiming*) arg;

        if (!this->quit) {

                // measuring period nanoseconds

                static struct timespec ts_new;


                clock_gettime(CLOCK_MONOTONIC, &ts_new);
                if (this->running) {
                        long int diff = ts_new.tv_nsec - this->ts_old.tv_nsec;
                        diff = diff<0 ? diff+1E9 : diff;

                        float cycles = 48000.0*diff/32.0/1E9;

                        if (this->verbose>=0 && cycles >= 2.0 )
                                rt_msg_buf_printf(this->buf,
                                "cycle_took_longer_than_expected %lu %f\n",
                                                  diff,
                                                  cycles);

                        if (this->verbose>=1)
                                rt_msg_buf_printf(this->buf,
                                "period_ns: %d %f\n",
                                          diff,
                                          cycles);

                        if (this->verbose>=2)
                        rt_msg_buf_printf(this->buf,"time: %lu.%09lu %f\n",
                                          ts_new.tv_sec,
                                          ts_new.tv_nsec);

                }
                this->ts_old = ts_new;
                this->running = 1;


        }


        return 0;
}






JackTiming *state_ptr = NULL;






void signal_handler (int ignore_this_arg)
{
        (void) ignore_this_arg;
        state_ptr->quit = 1;
}






void error (const char *desc)
{
        fprintf (stderr, "JACK error: %s\n", desc);
}






void jack_shutdown (void *arg)
{
        JackTiming  *this = (JackTiming*) arg; 
        fprintf(stderr,"Jack shut us down!\n");
        this->quit = 1;
}






void print_usage(char *name, FILE *to)
{
        /* I lost 20% in a university unix assignment once because I spelt it
         * "usage" instead of "Usage".  NEVER AGAIN!
         */
        fprintf(to,
                "%s "JACK_TIMING_VERSION": Gathers timing info from a pair "
                    "of jack servers.\n",name);
        fprintf(to,
                "    (Author: Dave Cunningham <spark@xdev.net>  License:GPL)\n"
                "\n"
                "Usage: %s [-h] [-v] [-q]\n"
                "\n"
                "        -h --help\tDisplay this usage info\n"
                "        -v --verbose\tMore status messages\n"
                "        -q --quiet\tFewer messages (except fatal errors)\n"
                "\n"
                ,name);
}






int main (int argc, char *argv[])
{

        char *app_name = strrchr(argv[0], '/');
        app_name = app_name==NULL ? argv[0] : app_name + 1;

        /* test that NULL is infact 0 */
        void *ptr;
        memset(&ptr,0,sizeof(ptr));
        if (ptr!=NULL) {
                fprintf(stderr,"This program assumes the NULL pointer is 0\n");
                exit(EXIT_FAILURE);
        }

        JackTiming state;
        memset(&state,0,sizeof(state)); /* set all initial pointers to NULL */
        state_ptr = &state; /* global variable for signal handler */

        if(signal(SIGINT,signal_handler)==SIG_ERR) {
                perror("Registering SIGINT handler.");
                exit(EXIT_FAILURE);
        }

        const char *options = "hvq" "N:1:2:";

        struct option long_options[] = {
                {"help", 0, 0, 'h'},
                {"verbose", 0, 0, 'v'},
                {"quiet", 0, 0, 'q'},

                {0, 0, 0, 0}
        };

        int opt = getopt_long(argc, argv, options, long_options, NULL);

        int verbose = 0;

        char *name = app_name;

        while (opt!=EOF) {
                switch (opt) {
                case 'h':
                        print_usage(app_name,stdout);
                        return EXIT_SUCCESS;
                case 'v':
                        verbose++;
                        break;
                case 'q':
                        verbose--;
                        break;
                default:
                        fprintf(stderr, "Unknown option: %c\n", optopt);
                        print_usage(app_name,stderr);
                        return EXIT_FAILURE;
                }
                opt = getopt_long(argc, argv, options, long_options, NULL);
        }

        int error = mlockall(MCL_CURRENT || MCL_FUTURE);
        if (error) {
                jack_timing_del(&state);
                fprintf(stderr,"Failed to mlock memory %s.\n",app_name);
                exit(EXIT_FAILURE);
        }

        error = jack_timing_init(&state, name, verbose);
        if (error) {
                jack_timing_del(&state);
                fprintf(stderr,"Failed to initialise %s.\n",app_name);
                exit(EXIT_FAILURE);
        }

        jack_activate(state.server);


        struct timespec ts = {0, 5E7};

        while (!state.quit) {
                nanosleep (&ts,0);
                rt_msg_buf_poll(state.buf,stdout);
                fflush(stdout);
                fflush(stderr);
        }

        rt_msg_buf_poll(state.buf,stdout);

        jack_timing_del (&state);

        return EXIT_SUCCESS;
}
